/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CPU_OPERATOR_SCALINGIMPL_KERNELS_H_
#define AIDGE_CPU_OPERATOR_SCALINGIMPL_KERNELS_H_

#include <cmath>
#include <cstddef>
#include "aidge/utils/Registrar.hpp"
#include "aidge/backend/cpu/operator/ScalingImpl.hpp"

//TODO : improve propagate, n2d2 :
/*
template<typename T>
void N2D2::floatingPointScaling_propagate(const Tensor<T>& input, Tensor<T>& output,
                                          std::size_t batchSize, std::size_t nbChannels,
                                          std::size_t height, std::size_t width,
                                          bool isClipped,
                                          const std::vector<Float_T>& clippingFactorPerChannel,
                                          const std::vector<Float_T>& scalingFactorPerChannel,
                                          std::size_t quantizedNbBits, bool isOutputUnsigned)
{
    std::size_t index = 0;
    for (std::size_t batch = 0; batch < batchSize; batch++) {
        for(std::size_t ch = 0; ch < nbChannels; ch++) {
            for(std::size_t y = 0; y < height; y++) {
                for(std::size_t x = 0; x < width; x++) {

                    T res = isClipped ? Clip(input(index), clippingFactorPerChannel[ch])
                                    : input(index);
                    res = Scale(res, scalingFactorPerChannel[ch]);

                    if(quantizedNbBits > 0) {
                        res = saturate(std::round(res), quantizedNbBits, isOutputUnsigned);
                    }
                    output(index) = (T) res;
                    index++;
                }
            }
        }
    }
}
*/


namespace Aidge {

template <class O>
const O& clamp(const O& x, const O& min, const O& max)
{
    return (x < min) ? min : (x > max) ? max : x;
}

template<class O>
O saturate(const O value, const std::size_t quantizedNbBits, const bool isOutputUnsigned) {
    // TODO: no assertions in kernel
    assert(quantizedNbBits > 0);

    const O min = isOutputUnsigned ? 0 :
                                  -(1ll << (quantizedNbBits - 1ll));
    const O max = isOutputUnsigned ? (1ll << quantizedNbBits) - 1ll :
                                   (1ll << (quantizedNbBits - 1ll)) - 1ll;

    return clamp(value, min, max);
}

template <class I, class O>
void ScalingImpl_cpu_forward_kernel(const float scalingFactor,
                                    const std::size_t quantizedNbBits,
                                    const bool isOutputUnsigned,
                                    std::size_t inputLenght,
                                    const void* input_,
                                    void* output_) {

    const I* input = static_cast<const I*>(input_);
    O* output = static_cast<O*>(output_);

    for (std::size_t i = 0; i < inputLenght; ++i) {
        output[i] = static_cast<O>(input[i] * static_cast<I>(scalingFactor));

        if(quantizedNbBits > 0) {
            output[i] = saturate(std::round(output[i]), quantizedNbBits, isOutputUnsigned);
        }
    }
}

// Kernels registration to implementation entry point
REGISTRAR(ScalingImpl_cpu,
    {DataType::Float32},
    {ProdConso::inPlaceModel, Aidge::ScalingImpl_cpu_forward_kernel<float, float>, nullptr});
REGISTRAR(ScalingImpl_cpu,
    {DataType::Float64},
    {ProdConso::inPlaceModel, Aidge::ScalingImpl_cpu_forward_kernel<double, double>, nullptr});
REGISTRAR(ScalingImpl_cpu,
    {DataType::Int32},
    {ProdConso::inPlaceModel, Aidge::ScalingImpl_cpu_forward_kernel<int32_t, int32_t>, nullptr});
}  // namespace Aidge

#endif /* AIDGE_CPU_OPERATOR_SCALINGIMPL_KERNELS_H_ */